"""Nodes of the Query `Tree`."""

from collections.abc import Callable, Container, Hashable
from dataclasses import KW_ONLY, dataclass, field
from enum import IntEnum, auto, unique
from typing import Any, Generic, Protocol

from container_data_collector.common_typevars import Inner, Outer
from container_data_collector.context import Context


@unique
class State(IntEnum):
    """State at which node processing is completed."""
    SUCCESS = auto()
    REJECT = auto()


@dataclass(slots=True)
class Result(Generic[Outer, Inner]):
    """Result of node processing."""
    state: State


class Node(Protocol):
    """Atomic unit of the Query `Tree`. Each node takes an object processed by
    previous node and the current context to know how process this object.
    The `Result` of the processing has some `State`, that can trigger some events
    in the previous node.
    """
    def process(self, obj: Any, context: Context[Outer, Inner]) -> Result[Outer, Inner]:
        ...

    def connect_with(self, *args: "Node") -> None:
        ...


@dataclass(slots=True, eq=False, match_args=False)
class PseudoRoot:
    """Pseudo root `Node` of the Query `Tree`."""
    next_nodes: list[Node] = field(default_factory=list, init=False)

    def process(self, obj: Any, context: Context[Outer, Inner]) -> Result[Outer, Inner]:
        for node in self.next_nodes:
            result = node.process(obj, context)
            match result.state:
                case State.SUCCESS:
                    continue
                case State.REJECT:
                    return result
        return Result(State.SUCCESS)

    def connect_with(self, *args: Node) -> None:
        self.next_nodes.extend(args)


@dataclass(slots=True, eq=False, match_args=False)
class Element:
    """`Node` that represents one of the elements collecting during tree traversal.
    Position tells the inserter function what position the element has in it.
    Next node may point to additional filters, e.g., can this element be used or not.
    """
    pos: int
    _: KW_ONLY
    next_node: Node | None = field(default=None, init=False)

    def __post_init__(self) -> None:
        if self.pos < 1:
            msg = "The position of the element must be greater than zero."
            hint = f"Instead, pos={self.pos} is given."
            raise ValueError(msg + "\n" + hint)

    def process(self, obj: Any, context: Context[Outer, Inner]) -> Result[Outer, Inner]:
        if self.next_node:
            result = self.next_node.process(obj, context)
            match result.state:
                case State.SUCCESS:
                    pass
                case State.REJECT:
                    return result

        context.apply_element(obj, pos=self.pos)
        return Result(State.SUCCESS)

    def connect_with(self, *args: Node) -> None:
        if len(args) != 1:
            raise ValueError("The Element node can have only one node.")
        self.next_node = args[0]


@dataclass(slots=True, eq=False, match_args=False)
class Group:
    """`Node` that represents one of the keys by which the elements will be grouped.
    Level tells what level of grouping this key is related. Factory tells how
    to make the key hashable if it is not. Next node may point to additional
    filters, e.g., can this key be used or not.
    """
    level: int
    _: KW_ONLY
    factory: Callable[[Any], Hashable] | None = field(default=None)
    next_node: Node | None = field(default=None, init=False)

    def __post_init__(self) -> None:
        if self.level < 1:
            msg = "The level of the group must be greater than zero."
            hint = f"Instead, level={self.level} is given."
            raise ValueError(msg + "\n" + hint)

    def process(self, obj: Any, context: Context[Outer, Inner]) -> Result[Outer, Inner]:
        if self.factory is not None:
            obj = self.factory(obj)
        if self.next_node:
            result = self.next_node.process(obj, context)
            match result.state:
                case State.SUCCESS:
                    pass
                case State.REJECT:
                    return result

        context.create_group(obj, pos=self.level)
        return Result(State.SUCCESS)

    def connect_with(self, *args: Node) -> None:
        if len(args) != 1:
            raise ValueError("The Group node can have only one node.")
        self.next_node = args[0]


@dataclass(slots=True, eq=False, match_args=False, kw_only=True)
class Include:
    """`Node` that takes an object from the previous node and checks it.
    If it is not contained in the `any_of` container or it doesn't pass
    the check of the `validator` `Callable` object, the whole branch until
    the closest `ForEach` node will be rejected.

    If both `any_of` and `validator` are not defined, the node processing
    has no effect.
    """
    any_of: Container[Any] | None = field(default=None)
    validator: Callable[[Any], bool] | None = field(default=None)

    def process(self, obj: Any, context: Context[Outer, Inner]) -> Result[Outer, Inner]:
        if self.any_of is not None and obj not in self.any_of:
            return Result(State.REJECT)
        if self.validator is not None and not self.validator(obj):
            return Result(State.REJECT)
        return Result(State.SUCCESS)

    def connect_with(self, *args: Node) -> None:
        class_name = self.__class__.__name__
        msg = f"The terminal node '{class_name}' is a leaf, so it can't have any attached nodes."
        raise NotImplementedError(msg)


@dataclass(slots=True, eq=False, match_args=False, kw_only=True)
class Exclude:
    """`Node` that takes an object from the previous node and checks it.
    If it is contained in the `any_of` container or it passes the check of
    the `invalidator` `Callable` object, the whole branch until the closest
    `ForEach` node will be rejected.

    If both `any_of` and `invalidator` are not defined, the node processing
    has no effect.
    """
    any_of: Container[Any] | None = field(default=None)
    invalidator: Callable[[Any], bool] | None = field(default=None)

    def process(self, obj: Any, context: Context[Outer, Inner]) -> Result[Outer, Inner]:
        if self.any_of is not None and obj in self.any_of:
            return Result(State.REJECT)
        if self.invalidator is not None and self.invalidator(obj):
            return Result(State.REJECT)
        return Result(State.SUCCESS)

    def connect_with(self, *args: Node) -> None:
        class_name = self.__class__.__name__
        msg = f"The terminal node '{class_name}' is a leaf, so it can't have any attached nodes."
        raise NotImplementedError(msg)


@dataclass(slots=True, eq=False, match_args=False, kw_only=True)
class At:
    """`Node` that takes an object from the previous node and treats it as
    a `Mapping` object, gets the value by key and propogates this value to
    the next nodes.
    """
    next_nodes: list[Node] = field(default_factory=list, init=False)
    key: Hashable

    def process(self, obj: Any, context: Context[Outer, Inner]) -> Result[Outer, Inner]:
        value = obj[self.key]
        for node in self.next_nodes:
            result = node.process(value, context)
            match result.state:
                case State.SUCCESS:
                    continue
                case State.REJECT:
                    return result
        return Result(State.SUCCESS)

    def connect_with(self, *args: Node) -> None:
        self.next_nodes.extend(args)


@dataclass(slots=True, eq=False, match_args=False, kw_only=True)
class ForEach:
    """`Node` that takes an object from the previous node and treats it
    as an `Iterable` object, iterates over it, and propogates each result
    to the next nodes.
    """
    next_nodes: list[Node] = field(default_factory=list, init=False)

    def process(self, obj: Any, context: Context[Outer, Inner]) -> Result[Outer, Inner]:
        for value in obj:
            for node in self.next_nodes:
                node.process(value, context)
        return Result(State.SUCCESS)

    def connect_with(self, *args: Node) -> None:
        self.next_nodes.extend(args)



"""`VectorPartial` class that describes a simplified partial object with only
non-default positional arguments and only default keyword arguments.
"""

import inspect
from enum import IntEnum, auto
from dataclasses import dataclass
from collections.abc import Callable
from typing import Any, Concatenate, Generic, Self, TypeAlias, TypeVar, final


@final
class _ArgState(IntEnum):
    """Stub saying that there is no argument."""
    NONE = auto()


FixedT_contra = TypeVar("FixedT_contra", contravariant=True)
ReturnT_co = TypeVar("ReturnT_co", covariant=True)

FuncWithFixedArgType: TypeAlias = Callable[Concatenate[FixedT_contra, ...], ReturnT_co]

@final
@dataclass(slots=True, eq=False, match_args=False)
class VectorPartial(Generic[FixedT_contra, ReturnT_co]):
    """Partial wrapper for functions with the fixed type of the first argument
    and the return value type. The arguments of the functions can be added and
    removed at the specified position in constant time. After all the argument
    was added, the function can be called, but if someone of the arguments is
    missing, raise `RuntimeError`.

    These function must have at least one positional argument, can't have any
    positional arguments with a default value, and can't have any keyword
    arguments without a default value. Raise `TypeError` if someone of these
    conditions is not complited.

    Passed the `n_args` argument describes how many positional arguments
    the function has to take. Raise `IndexError` if it is not greater than zero.
    Raise `TypeError` if it doesn't match with the number of the function's
    arguments.
    """

    _func: FuncWithFixedArgType[FixedT_contra, ReturnT_co]
    _args: list[Any]
    _missing_args: int

    def __init__(self, func: FuncWithFixedArgType[FixedT_contra, ReturnT_co],
                 /, *,
                 n_args: int) -> None:
        VectorPartial._init_check(func, n_args=n_args)
        self._func = func
        self._args = [_ArgState.NONE for _ in range(n_args)]
        self._missing_args = n_args

    def __call__(self) -> ReturnT_co:
        if not self.can_return:
            msg = f"The function {self._func.__name__} can't be called:"
            hint = f"it is still waiting {self.missing_args} arguments."
            raise RuntimeError(msg + " " + hint)
        return self._func(*self._args)

    def __copy__(self) -> Self:
        cls = self.__class__
        shallow_copy = cls.__new__(cls)
        shallow_copy._func = self._func
        shallow_copy._args = self._args.copy()
        shallow_copy._missing_args = self._missing_args
        return shallow_copy

    @property
    def args_count(self) -> int:
        """The number of the arguments."""
        return len(self._args)

    @property
    def can_return(self) -> bool:
        """Return `True` if the function is ready to be called."""
        return self._missing_args == 0

    @property
    def missing_args(self) -> int:
        """Return the number of the arguments the function is waiting."""
        return self._missing_args

    def insert(self, arg: Any, /, *, pos: int) -> None:
        """Insert a new argument of the function at the specified position.
        Raise `IndexError` if 'pos' is not greater than zero and less or equal
        than a number of the arguments that the function expects.
        """
        self._check_pos(pos)
        index = pos - 1
        if self._args[index] is _ArgState.NONE:
            self._missing_args -= 1
        self._args[index] = arg

    def remove(self, pos: int) -> None:
        """Remove an argument at the specified position.
        Raise `IndexError` if 'pos' is not greater than zero and less or equal
        than a number of the arguments that the function expects.
        """
        self._check_pos(pos)
        index = pos - 1
        if self._args[index] is not _ArgState.NONE:
            self._args[index] = _ArgState.NONE
            self._missing_args += 1

    def _check_pos(self, pos: int) -> None:
        if not 0 < pos <= len(self._args):
            msg = f"Constraint 0 < pos <= {len(self._args)} is not satisfied.\n"
            hint = f"Instead, {pos=} is given."
            raise IndexError(msg + hint)

    @staticmethod
    def _init_check(func: FuncWithFixedArgType[FixedT_contra, ReturnT_co],
                    /, *,
                    n_args: int) -> None:
        if n_args < 1:
            message = f"n_args must be greater than zero.\nInstead, {n_args=} is given."
            raise IndexError(message)

        spec = inspect.getfullargspec(func)
        msg_head = f"The function '{func.__name__}'"
        if not spec.args:
            msg = f"{msg_head} must have at least one positional argument."
            raise TypeError(msg)
        if spec.defaults:
            msg = f"{msg_head} can't have any default positional arguments."
            raise TypeError(msg)
        if spec.kwonlyargs and len(spec.kwonlyargs) != len(spec.kwonlydefaults or []):
            msg = f"{msg_head} can't have non-default keyword arguments."
            raise TypeError(msg)
        if spec.varargs and n_args < len(spec.args):
            msg = f"{msg_head} must have at least {n_args=} positional arguments."
            hint = f"Instead, {len(spec.args)} is given."
            raise TypeError(msg + "\n" + hint)
        if not spec.varargs and n_args != len(spec.args):
            msg = f"{msg_head} must have exactly {n_args=} arguments."
            hint = f"Instead, {len(spec.args)} is given."
            raise TypeError(msg + "\n" + hint)



"""Collecting and grouping data from various containers."""

from container_data_collector.query import Query, Branch
from container_data_collector.collector import PlainCollector, GroupCollector

__all__ = [
    "Query",
    "Branch",
    "PlainCollector",
    "GroupCollector",
]



"""The Query `Tree`.

A simplified BNF of the Query `Tree` is presented below. (It only doesn't
specify what parameters the functions have and the correct number of these
functions.):

```
QueryTree ::= Element | At | ForEach | Branches

At ::= AtCall "." AtBranch
AtCall ::= "at(...)"
AtBranch ::= Group | Filter | QueryTree

ForEach ::= ForEachCall "." QueryTree
ForEachCall ::= "for_each()"

Branches ::= "branches(" BranchList ")"
BranchList ::= BranchListWithoutForEach
                | At "," AtOnlyBranchList
                | AtOnlyBranchList "," At "," AtOnlyBranchList
                | AtOnlyBranchList "," At
AtOnlyBranchList ::= AtOnly
                    | AtOnly "," AtOnlyBranchList

AtOnly ::= AtCall "." AtOnlyBranch
AtOnlyBranch ::= Element | Group | Filter | AtOnly | BranchesWithoutForEach

BranchesWithoutForEach ::= "branches(" BranchListWithoutForEach ")"
BranchListWithoutForEach ::= AtOnlyBranchList "," AtOnlyBranchList

Filter ::= Include | Exclude
Include ::= "include(...)"
Exclude ::= "exclude(...)"

Element ::= ElementCall | ElementCall "." Filter
ElementCall ::= "element(...)"

Group ::= GroupCall | GroupCall "." Filter
GroupCall ::= "group(...)"
```
"""

from collections.abc import Callable, Hashable
from dataclasses import KW_ONLY, dataclass, field
from enum import IntFlag, unique
from typing import Any, Container, Generic, Self, TypeAlias, TypeVar, final

from container_data_collector.nodes import (
    At,
    Element,
    Exclude,
    ForEach,
    Group,
    Include,
    Node,
    PseudoRoot,
)


@final
@dataclass(slots=True, eq=False, match_args=False)
class Tree:
    """Query `Tree` that represents a query by which collecting and grouping of
    elements takes place.
    """
    root: Node
    _: KW_ONLY
    n_elements: int
    n_groups: int


@final
@unique
class _TreePriority(IntFlag):
    """Key by which subtrees are sorted in a list of branches."""
    DEFAULT = 0
    FILTER = 1
    GROUP = 2
    ELEMENT = 4
    FOR_EACH = 8


@dataclass(slots=True, eq=False, match_args=False)
class TreeFactory:
    """Factory that generate a query tree from the current query chain builder
    and the upstream query chain builder, if the latest one exists.
    """
    _builder: "_ChainBuilder"

    def __call__(self, upstream_chain: "_ChainBuilder | None" = None) -> Tree:
        if not upstream_chain:
            return self._builder.build()
        upstream_chain.connect_with(self._builder)
        return upstream_chain.build()


@final
class _Trunk:
    pass

@final
class _Branch:
    pass

_TreePartT = TypeVar("_TreePartT", _Trunk, _Branch)

class _Terminal(TreeFactory, Generic[_TreePartT]):
    def __lt__(self, rhs: Self) -> bool:
        return self._builder.priority < rhs._builder.priority

    @classmethod
    def connect_branches(cls, upstream_chain: "_ChainBuilder",
                         first: Self, second: Self, *args: Self) -> None:
        """Connect the current node of the `upstream_chain` builder with passed
        `_Terminal` branches.
        
        Raise `RuntimeError`, if at least two branches have a `ForEach` node.
        """
        factories = [first, second, *args]
        factories.sort()
        if factories[-2]._builder.priority >= _TreePriority.FOR_EACH:
            msg = "There cannot be ForEach nodes in different branches simultaneously."
            raise RuntimeError(msg)

        upstream_chain.connect_with(*(factory._builder for factory in factories))


@final
@dataclass(slots=True, eq=False, match_args=False, kw_only=True)
class _ChainBuilder:
    _root: Node
    _current: Node
    _priority: _TreePriority = field(default=_TreePriority.DEFAULT)
    _elements: list[int] = field(default_factory=list)
    _groups: list[int] = field(default_factory=list)

    @classmethod
    def create(cls) -> Self:
        """Create the builder with a `PseudoRoot` node in the root."""
        root = PseudoRoot()
        return cls(_root=root, _current=root)

    @property
    def root(self) -> Node:
        """Return the root of the builder."""
        return self._root

    @property
    def priority(self) -> _TreePriority:
        """Return the current priority of the builder."""
        return self._priority

    @property
    def elements(self) -> list[int]:
        """Return a list of the positions of the elements."""
        return self._elements

    @property
    def groups(self) -> list[int]:
        """Return a list of the positions of the groups."""
        return self._groups

    def build(self) -> Tree:
        """Build the completed Tree.
        
        The `Tree` must have at least one `Element` node. Positions of `Element` and
        `Group` nodes, if the last ones are presented, must have end-to-end
        numbering and start with one. Otherwise, `RuntimeError` will be raised.
        """
        _check_positions_order(self._groups, "groups")
        _check_positions_order(self._elements, "elements")
        if not self._elements:
            msg = "There must be at leats one element in the query tree."
            raise RuntimeError(msg)

        return Tree(self._root, n_elements=len(self._elements), n_groups=len(self._groups))

    def connect_with(self, *args: Self) -> None:
        """Connect the current builder's node with the roots of the passed
        builders, and also update its other properties. 
        """
        if not args:
            return None
        if len(args) == 1:
            builder = args[0]
            self._priority |= builder.priority
            self._elements.extend(builder.elements)
            self._groups.extend(builder.groups)
            return None

        nodes = list[Node]()
        priority = _TreePriority.DEFAULT
        elements = list[int]()
        groups = list[int]()
        for builder in args:
            nodes.append(builder.root)
            priority |= builder.priority
            elements.extend(builder.elements)
            groups.extend(builder.groups)

        self._current.connect_with(*nodes)
        self._priority |= priority
        self._elements.extend(elements)
        self._groups.extend(groups)
        return None

    def element(self, pos: int = 1) -> None:
        """Create `Element` node and add it to the end of the builder chain."""
        node = Element(pos)
        self._current.connect_with(node)
        self._current = node
        self._priority |= _TreePriority.ELEMENT
        self._elements.append(pos)

    def group(self, level: int = 1, /, *, factory: Callable[[Any], Hashable] | None = None) -> None:
        """Create `Group` node and add it to the end of the builder chain."""
        node = Group(level, factory=factory)
        self._current.connect_with(node)
        self._current = node
        self._priority |= _TreePriority.GROUP
        self._groups.append(level)

    def include(self, /, *,
                any_of: Container[Any] | None = None,
                validator: Callable[[Any], bool] | None = None) -> None:
        """Create `Include` node and add it to the end of the builder chain."""
        node = Include(any_of=any_of, validator=validator)
        self._current.connect_with(node)
        self._current = node
        self._priority |= _TreePriority.FILTER

    def exclude(self, /, *,
                any_of: Container[Any] | None = None,
                invalidator: Callable[[Any], bool] | None = None) -> None:
        """Create `Exclude` node and add it to the end of the builder chain."""
        node = Exclude(any_of=any_of, invalidator=invalidator)
        self._current.connect_with(node)
        self._current = node
        self._priority |= _TreePriority.FILTER

    def at(self, key: Hashable) -> None:
        """Create `At` node and add it to the end of the builder chain."""
        node = At(key=key)
        self._current.connect_with(node)
        self._current = node

    def for_each(self) -> None:
        """Create `ForEach` node and add it to the end of the builder chain."""
        node = ForEach()
        self._current.connect_with(node)
        self._current = node
        self._priority |= _TreePriority.FOR_EACH

    def branches(self,
                 first: _Terminal[_Branch],
                 second: _Terminal[_Branch],
                 *args: _Terminal[_Branch]) -> None:
        """Connect the current builder's node with the roots of the passed
        branches."""
        _Terminal.connect_branches(self, first, second, *args)


def _check_positions_order(xs: list[int], xs_name: str) -> None:
    xs.sort()
    pos = 1
    missing_intervals = list[tuple[int, int]]()
    missing_positions = 0
    for x in xs:
        if x != pos:
            missing_intervals.append((pos, x - 1))
            missing_positions += x - pos
            pos = x
        pos += 1

    if missing_intervals:
        verb, s = ("is", "") if missing_positions == 1 else ("are", "s")
        msg = f"In the {xs_name}: there {verb} {missing_positions} missing position{s}."
        missings = [
            (
                str(interval[0])
                if interval[0] == interval[1]
                else f"{interval[0]}..{interval[1]}"
            )
            for interval in missing_intervals
        ]
        hint = f"Missing: {', '.join(missings)}."
        raise RuntimeError(msg + "\n" + hint)


@final
class _WithFilter(_Terminal[_TreePartT]):
    def include(self, /, *,
                any_of: Container[Any] | None = None,
                validator: Callable[[Any], bool] | None = None
                ) -> _Terminal[_TreePartT]:
        """The Query `Tree` node that correspondes to a filter, that takes
        an object, and checks if it is contained in the `any_of` container or
        it passes the `validator`. Otherwise, the whole branch until the closest
        ForEach node will be rejected.

        It always loses against `exclude` if it processes the same object.

        This method is one of the terminal methods of the Query `Tree` builder.
        """
        self._builder.include(any_of=any_of, validator=validator)
        return _Terminal(self._builder)

    def exclude(self, /, *,
                any_of: Container[Any] | None = None,
                invalidator: Callable[[Any], bool] | None = None
                ) -> _Terminal[_TreePartT]:
        """The Query `Tree` node that correspondes to a filter, that takes
        an object, and checks if it is not contained in the `any_of` container
        or it doesn't pass the `invalidator`. Otherwise, the whole branch until
        the closest ForEach node will be rejected.

        It always wins against `include` if it processes the same object.

        This method is one of the terminal methods of the Query `Tree` builder.
        """
        self._builder.exclude(any_of=any_of, invalidator=invalidator)
        return _Terminal(self._builder)


@final
class _AtChain(TreeFactory, Generic[_TreePartT]):
    def element(self, pos: int = 1) -> _WithFilter[_TreePartT]:
        """The Query `Tree` node that corresponds to getting the current value
        and applying it to the inserter partial function by the specified
        position.
        
        The position must be greater than zero, and if in the Query
        `Tree`, there are two elements: with position `1` and with the largest
        position `N`, then also there must be elements with position `2`, `3`,
        and so on up to `N-1`.

        This method can be treated as the terminal method of the Query `Tree`
        builder.

        Possible next methods: `include`, `exclude`.
        """
        self._builder.element(pos)
        return _WithFilter(self._builder)

    def group(self, level: int = 1,
              /, *,
              factory: Callable[[Any], Hashable] | None = None
              ) -> _WithFilter[_TreePartT]:
        """The Query `Tree` node that corresponds to getting the current value
        and grouping collecting elements by it. The level of the group is
        determined by value of the `level`.
        
        The `level` must be greater than zero, and if in the Query
        `Tree`, there are two groups: with level `1` and with the largest
        level `N`, then also there must be groups with level `2`, `3`,
        and so on up to `N-1`.

        This method can be treated as the terminal method of the Query `Tree`
        builder.

        Possible next methods: `include`, `exclude`.
        """
        self._builder.group(level, factory=factory)
        return _WithFilter(self._builder)

    def include(self, /, *,
                any_of: Container[Any] | None = None,
                validator: Callable[[Any], bool] | None = None
                ) -> _Terminal[_TreePartT]:
        """The Query `Tree` node that correspondes to a filter, that takes
        an object, and checks if it is contained in the `any_of` container or
        it passes the `validator`. Otherwise, the whole branch until the closest
        ForEach node will be rejected.

        It always loses against `exclude` if it processes the same object.

        This method is one of the terminal methods of the Query `Tree` builder.
        """
        self._builder.include(any_of=any_of, validator=validator)
        return _Terminal(self._builder)

    def exclude(self, /, *,
                any_of: Container[Any] | None = None,
                invalidator: Callable[[Any], bool] | None = None
                ) -> _Terminal[_TreePartT]:
        """The Query `Tree` node that correspondes to a filter, that takes
        an object, and checks if it is not contained in the `any_of` container
        or it doesn't pass the `invalidator`. Otherwise, the whole branch until
        the closest ForEach node will be rejected.

        It always wins against `include` if it processes the same object.

        This method is one of the terminal methods of the Query `Tree` builder.
        """
        self._builder.exclude(any_of=any_of, invalidator=invalidator)
        return _Terminal(self._builder)

    def at(self, key: Hashable) -> Self:
        """The Query `Tree` node that corresponds to a lookup by key in
        a `Mapping` container.

        Possible next methods: `element`, `group`, `include`, `exclude`, `at`,
        `for_each`, `branches`.
        """
        self._builder.at(key)
        return _AtChain(self._builder)

    def for_each(self) -> "_ForEachChain[_TreePartT]":
        """The Query `Tree` node that corresponds to an `Iterable` object, therefore,
        when getting subobjects by iteration over this `Iterable` object, all these
        subobjects will be processed one-by-one by the next node.

        Possible next methods: `element`, `at`, `for_each`, `branches`.
        """
        self._builder.for_each()
        return _ForEachChain(self._builder)

    def branches(self,
                 first: _Terminal[_Branch],
                 second: _Terminal[_Branch],
                 *args: _Terminal[_Branch]
                 ) -> _Terminal[_Branch]:
        """The way to make two or more branches in the Query `Tree`. The order, in
        which these branches list, is not important, but there shouldn't be two
        `for_each` in the different branches. All branches have to be started
        with `Branch.at(<some hashable key>)`.

        This method is one of the terminal methods of the Query `Tree` builder.
        """
        self._builder.branches(first, second, *args)
        return _Terminal(self._builder)


@final
class _ForEachChain(TreeFactory, Generic[_TreePartT]):
    def element(self, pos: int = 1) -> _WithFilter[_TreePartT]:
        """The Query `Tree` node that corresponds to getting the current value
        and applying it to the inserter partial function by the specified
        position.
        
        The position must be greater than zero, and if in the Query
        `Tree`, there are two elements: with position `1` and with the largest
        position `N`, then also there must be elements with position `2`, `3`,
        and so on up to `N-1`.

        This method can be treated as the terminal method of the Query `Tree`
        builder.

        Possible next methods: `include`, `exclude`.
        """
        self._builder.element(pos)
        return _WithFilter(self._builder)

    def at(self, key: Hashable) -> _AtChain[_TreePartT]:
        """The Query `Tree` node that corresponds to a lookup by key in
        a `Mapping` container.

        Possible next methods: `element`, `group`, `include`, `exclude`, `at`,
        `for_each`, `branches`.
        """
        self._builder.at(key)
        return _AtChain(self._builder)

    def for_each(self) -> Self:
        """The Query `Tree` node that corresponds to an `Iterable` object, therefore,
        when getting subobjects by iteration over this `Iterable` object, all these
        subobjects will be processed one-by-one by the next node.

        Possible next methods: `element`, `at`, `for_each`, `branches`.
        """
        self._builder.for_each()
        return _ForEachChain(self._builder)

    def branches(self,
                 first: _Terminal[_Branch],
                 second: _Terminal[_Branch],
                 *args: _Terminal[_Branch]
                 ) -> _Terminal[_Branch]:
        """The way to make two or more branches in the Query `Tree`. The order, in
        which these branches list, is not important, but there shouldn't be two
        `for_each` in the different branches. All branches have to be started
        with `Branch.at(<some hashable key>)`.

        This method is one of the terminal methods of the Query `Tree` builder.
        """
        self._builder.branches(first, second, *args)
        return _Terminal(self._builder)


class _ChainEntryPointBase(Generic[_TreePartT]):
    @classmethod
    def at(cls, key: Hashable) -> _AtChain[_TreePartT]:
        """The Query `Tree` node that corresponds to a lookup by key in
        a `Mapping` container.

        Possible next methods: `element`, `group`, `include`, `exclude`, `at`,
        `for_each`, `branches`.
        """
        builder = _ChainBuilder.create()
        builder.at(key)
        return _AtChain(builder)


@final
class _ChainEntryPoint(_ChainEntryPointBase[_TreePartT]):
    @classmethod
    def element(cls, pos: int = 1) -> _WithFilter[_Trunk]:
        """The Query `Tree` node that corresponds to getting the current value
        and applying it to the inserter partial function by the specified
        position.
        
        The position must be greater than zero, and if in the Query
        `Tree`, there are two elements: with position `1` and with the largest
        position `N`, then also there must be elements with position `2`, `3`,
        and so on up to `N-1`.

        This method can be treated as the terminal method of the Query `Tree`
        builder.

        Possible next methods: `include`, `exclude`.
        """
        builder = _ChainBuilder.create()
        builder.element(pos)
        return _WithFilter(builder)

    @classmethod
    def for_each(cls) -> _ForEachChain[_Trunk]:
        """The Query `Tree` node that corresponds to an `Iterable` object, therefore,
        when getting subobjects by iteration over this `Iterable` object, all these
        subobjects will be processed one-by-one by the next node.

        Possible next methods: `element`, `at`, `for_each`, `branches`.
        """
        builder = _ChainBuilder.create()
        builder.for_each()
        return _ForEachChain(builder)

    @classmethod
    def branches(cls,
                 first: _Terminal[_Branch],
                 second: _Terminal[_Branch],
                 *args: _Terminal[_Branch]) -> TreeFactory:
        """The way to make two or more branches in the Query `Tree`. The order, in
        which these branches list, is not important, but there shouldn't be two
        `for_each` in the different branches. All branches have to be started
        with `Branch.at(<some hashable key>)`.

        This method is one of the terminal methods of the Query `Tree` builder.
        """
        builder = _ChainBuilder.create()
        builder.branches(first, second, *args)
        return TreeFactory(builder)


Query: TypeAlias = _ChainEntryPoint[_Trunk]
Branch: TypeAlias = _ChainEntryPointBase[_Branch]



"""The `Collector` class and its factories for more convenient work."""

import copy
from collections.abc import Callable, Hashable, Iterable
from dataclasses import dataclass
from typing import Any, Concatenate, Generic, final

from container_data_collector.common_typevars import Inner, Inner_contra, Outer
from container_data_collector.context import Context
from container_data_collector.query import Tree, TreeFactory
from container_data_collector.vector_partial import VectorPartial


@dataclass(slots=True, eq=False, match_args=False)
class _Collector(Generic[Outer, Inner]):
    _query_tree: Tree
    _outer_factory: Callable[[], Outer]
    _element_inserter: VectorPartial[Inner, None]
    _group_factory: VectorPartial[Outer, Inner]

    def collect(self, objects: Iterable[Any]) -> Outer:
        """Take all elements of the `Iterable` object and collect it according to
        the structure of the query tree, then returning the result.
        """
        result = self._outer_factory()
        context = Context[Outer, Inner](
            top_container=result,
            inserter=copy.copy(self._element_inserter),
            group_factory=copy.copy(self._group_factory),
        )

        for obj in objects:
            self._query_tree.root.process(obj, context)
        return result


@final
class PlainCollector(Generic[Inner_contra]):
    """This `Collector` factory is for those collectors that do no need to work
    with groups. If the query tree doesn't have any group elements, this factory
    should be used.
    """

    @classmethod
    def compile(cls, query: TreeFactory,
                /, *,
                inner_factory: Callable[[], Inner],
                inserter: Callable[Concatenate[Inner_contra, ...], None],
                ) -> _Collector[Inner, Inner]:
        """Return the `Collector` instance.
        
        The `query` argument is a structure of the query tree. It describes
        how the `Collector` should traverse passed objects during collecting.
        The query tree must have zero or only one path from the root to some
        leaf, that has `for_each` node. The query tree must have at least
        one `element` node. Integer numbers, that passed into `element` and
        `group` nodes, must have end-to-end numbering and start with one, e.g.
        1, 2, 3, if there are only three positions. Otherwise, it raises
        'RuntimeError'.

        The query has to be started with `Query.<suitable method>`.

        The `inserter` argument is about how collecting elements have to be
        inserted. This `Callable` object must have the first argument as some
        container in which the elements will be inserted, and the rest of
        the arguments are the elements themselves, which positions have to be
        the same as in the `element` nodes. The `Callable` object can't have any
        default positional arguments and non-default keyword arguments.
        Otherwise, it raises `TypeError`.

        In the other hand, the `inner_factory` argument says how to get
        the container in which the elements will be inserted in the way
        the `inserter` describes it. It have to be a plain producer-like
        `Callable` object, that receives no arguments and returns the container.
        """
        def _group_factory_0(c: Inner) -> Inner:
            return c

        tree = query()
        element_inserter = VectorPartial(inserter, n_args=tree.n_elements+1)
        if tree.n_groups:
            msg = "There cannot be any groups in the query using ListCollector."
            hint = f"Instead, {tree.n_groups} {'is' if tree.n_groups == 1 else 'are'} given."
            raise ValueError(msg + "\n" + hint)

        group_factory = VectorPartial(_group_factory_0, n_args=1)
        return _Collector(tree, inner_factory, element_inserter, group_factory)


@final
class GroupCollector(Generic[Inner_contra]):
    """This `Collector` factory is for those collectors that do need to work with
    groups. If the query tree has any group elements, this factory should be used.
    """

    @classmethod
    def compile(cls, query: TreeFactory,
                /, *,
                inner_factory: Callable[[], Inner],
                inserter: Callable[Concatenate[Inner_contra, ...], None],
                ) -> _Collector[dict[Hashable, Any], Inner]:
        """Return the `Collector` instance.
        
        The `query` argument is a structure of the query tree. It describes
        how the `Collector` should traverse passed objects during collecting.
        The query tree must have zero or only one path from the root to some
        leaf, that has `for_each` node. The query tree must have at least
        one `element` node. Integer numbers, that passed into `element` and
        `group` nodes, must have end-to-end numbering and start with one, e.g.
        1, 2, 3, if there are only three positions. Otherwise, it raises
        'RuntimeError'.

        The query has to be started with `Query.<suitable method>`.

        The `inserter` argument is about how collecting elements have to be
        inserted. This `Callable` object must have the first argument as some
        container in which the elements will be inserted, and the rest of
        the arguments are the elements themselves, which positions have to be
        the same as in the `element` nodes. The `Callable` object can't have any
        default positional arguments and non-default keyword arguments.
        Otherwise, it raises `TypeError`.

        In the other hand, the `inner_factory` argument says how to get
        the container in which the elements will be inserted in the way
        the `inserter` describes it. It have to be a plain producer-like
        `Callable` object, that receives no arguments and returns the container.
        """
        def _group_factory_1(c: dict[Hashable, Any], key_1: Hashable) -> Inner:
            return c.setdefault(key_1, inner_factory())

        def _group_factory_2(c: dict[Hashable, Any], key_1: Hashable, key_2: Hashable) -> Inner:
            return c.setdefault(
                    key_1, {}
                ).setdefault(key_2, inner_factory())

        def _group_factory_3(c: dict[Hashable, Any],
                            key_1: Hashable, key_2: Hashable, key_3: Hashable) -> Inner:
            return c.setdefault(
                    key_1, {}
                ).setdefault(
                    key_2, {}
                ).setdefault(key_3, inner_factory())

        def _group_factory_4(c: dict[Hashable, Any],
                            key_1: Hashable, key_2: Hashable, key_3: Hashable,
                            key_4: Hashable) -> Inner:
            return c.setdefault(
                    key_1, {}
                ).setdefault(
                    key_2, {}
                ).setdefault(
                    key_3, {}
                ).setdefault(key_4, inner_factory())

        def _group_factory_5(c: dict[Hashable, Any],
                            key_1: Hashable, key_2: Hashable, key_3: Hashable,
                            key_4: Hashable, key_5: Hashable) -> Inner:
            return c.setdefault(
                    key_1, {}
                ).setdefault(
                    key_2, {}
                ).setdefault(
                    key_3, {}
                ).setdefault(
                    key_4, {}
                ).setdefault(key_5, inner_factory())

        def _group_factory_n(c: dict[Hashable, Any], *args: Hashable) -> Inner:
            assert len(args) > 5
            for key in args[:-1]:
                c = c.setdefault(key, {})
            return c.setdefault(args[-1], inner_factory())

        tree = query()
        element_inserter = VectorPartial(inserter, n_args=tree.n_elements+1)
        if not tree.n_groups:
            msg = "There must be at least one group in the query using GroupCollector."
            raise ValueError(msg)

        match tree.n_groups:
            case 1 as n:
                group_factory = VectorPartial(_group_factory_1, n_args=n+1)
            case 2 as n:
                group_factory = VectorPartial(_group_factory_2, n_args=n+1)
            case 3 as n:
                group_factory = VectorPartial(_group_factory_3, n_args=n+1)
            case 4 as n:
                group_factory = VectorPartial(_group_factory_4, n_args=n+1)
            case 5 as n:
                group_factory = VectorPartial(_group_factory_5, n_args=n+1)
            case n:
                group_factory = VectorPartial(_group_factory_n, n_args=n+1)

        return _Collector(tree, dict, element_inserter, group_factory)



"""Common `TypeVar`s for generic classes and functions.

`Inner` — a type for the inner container, in which elements are directly put.

`Outer` — a type for the outer container for collection and grouping elements.
If the number of groups is greater than zero, the `Outer` is a `Mapping` container
(usually a plain `dict`), according to the keys which the outermost grouping
occurs. Otherwise, the `Outer` is the `Inner`.
"""

from typing import TypeVar


Inner = TypeVar("Inner")
Inner_co = TypeVar("Inner_co", covariant=True)
Inner_contra = TypeVar("Inner_contra", contravariant=True)

Outer = TypeVar("Outer")
Outer_co = TypeVar("Outer_co", covariant=True)
Outer_contra = TypeVar("Outer_contra", contravariant=True)



"""Shared context for tree traversal."""

from collections.abc import Hashable
from dataclasses import dataclass
from typing import Any, Generic

from container_data_collector.common_typevars import Inner, Outer
from container_data_collector.vector_partial import VectorPartial


@dataclass(slots=True, eq=False, match_args=False)
class Context(Generic[Outer, Inner]):
    """Structure that joins two functions:
      - an element inserter;
      - a group factory.
    
    It describes the current state of collection of elements and grouping them
    during tree traversal.
    """

    _element_inserter: VectorPartial[Inner, None]
    _group_factory: VectorPartial[Outer, Inner]

    def __init__(self, /, *,
                 top_container: Outer,
                 inserter: VectorPartial[Inner, None],
                 group_factory: VectorPartial[Outer, Inner]) -> None:
        self._element_inserter = inserter
        self._group_factory = group_factory
        if group_factory.args_count == 1:
            self._element_inserter.insert(top_container, pos=1)
        else:
            self._group_factory.insert(top_container, pos=1)

    def apply_element(self, e: Any, /, *, pos: int) -> None:
        """Apply the passed element to the function-inserter. Then if
        the function can be called, it will be.

        Raise `ValueError`, if `pos` is 0.
        """
        Context._check_pos(pos)
        self._element_inserter.insert(e, pos=pos+1)
        if self._element_inserter.can_return:
            self._element_inserter()
            self._element_inserter.remove(pos + 1)

    def create_group(self, key: Hashable, /, *, pos: int) -> None:
        """Apply the passed key to the group factory function. Then if
        the function can be called, it will be, and the a new bottom container
        will be apply to the function-inserter as the first argument.

        Raise `ValueError`, if `pos` is 0.
        """
        Context._check_pos(pos)
        self._group_factory.insert(key, pos=pos+1)
        if self._group_factory.can_return:
            bottom_container = self._group_factory()
            self._group_factory.remove(pos + 1)
            self._element_inserter.insert(bottom_container, pos=1)

    @staticmethod
    def _check_pos(pos: int) -> None:
        if pos == 0:
            msg = "Manual applying a container to the function is not allowed."
            hint = f"{pos=} is forbidden."
            raise ValueError(msg + "\n" + hint)
